/*
 * IP_MASQ_MMS - Microsoft Messenger Service masquerading module for Linux
 * Author:  Tom Marshall <tommy@tig-grr.com>
 * Version: 0.91
 *
 *	 This program is free software; you can redistribute it and/or
 *	 modify it under the terms of the GNU General Public License
 *	 as published by the Free Software Foundation; either version
 *	 2 of the License, or (at your option) any later version.
 *
 * Revision History
 * ================
 * v0.91: Fixed bug in skbuff replacement (thanks for the repro David).
 *        Removed harmless printk warning.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/init.h>
#include <net/protocol.h>
#include <net/tcp.h>

#include <net/ip_masq.h>

/* The default port list */
static int ports[MAX_MASQ_APP_PORTS] =
{
    1863,
    0
}; 
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_MASQ_APP_PORTS) "i");

#define IS_DIGIT(c)    (c >= '0' && c <= '9')
#define IS_DIGITDOT(c) ((c >= '0' && c <= '9') || c == '.')
#define IS_ALPHA(c)    ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z'))
#define IS_WSPACE(c)   (c == ' ' || c == '\t' || c == '\r' || c == '\n')
#define SKIP_WSPACE(p,n) while( n > 0 && IS_WSPACE(*p) ) { n--; p++; }

static struct ip_masq_app* g_masq_objs[MAX_MASQ_APP_PORTS];

/*** helpers ***/

static int
str_to_u16( char* pbuf, __u16* pval )
{
    int n = 0;

    *pval = 0;
    while( pbuf[n] >= '0' && pbuf[n] <= '9' )
    {
        *pval = (*pval * 10) + (pbuf[n] - '0');
        n++;
    }

    return n;
}

static int
str_to_u32( char* pbuf, __u32* pval )
{
    int n = 0;

    *pval = 0;
    while( pbuf[n] >= '0' && pbuf[n] <= '9' )
    {
        *pval = (*pval * 10) + (pbuf[n] - '0');
        n++;
    }

    return n;
}

static int
next_line( char** ppbuf, __u32* pbuflen, char** ppline, __u32* plinelen )
{
    char* pbuf = *ppbuf;
    __u32 buflen = *pbuflen;
    __u32 linelen = 0;

    do
    {
        while( *pbuf != '\r' && *pbuf != '\n' )
        {
            if( buflen <= 1 )
            {
                return 0;
            }

            pbuf++;
            linelen++;
            buflen--;
        }

        if( buflen > 1 && *pbuf == '\r' && *(pbuf+1) == '\n' )
        {
            pbuf++;
            buflen--;
        }

        pbuf++;
        buflen--;
    }
    while( buflen > 0 && (*pbuf == ' ' || *pbuf == '\t') );

    *ppline = *ppbuf;
    *plinelen = linelen;
    *ppbuf = pbuf;
    *pbuflen = buflen;

    return linelen;
}

static void
get_skb_data( struct sk_buff** ppskb, char** ppdata, __u32* plen )
{
    struct sk_buff* skb;
    struct iphdr*   iph;
    struct tcphdr*  th;
    char* pbuf;
    char* pend;

    skb = *ppskb;
    iph = skb->nh.iph;
    th = (struct tcphdr*)&(((char*)iph)[iph->ihl*4]);

    pbuf = (char*)th + (th->doff * 4);
    pend = skb->h.raw + skb->len;
    *ppdata = pbuf;
    *plen = pend - pbuf;
}

/**************************************
 * masq funcs
 **************************************/

static int
masq_mms_init_1( struct ip_masq_app* pmapp, struct ip_masq* pms )
{
    MOD_INC_USE_COUNT;

    pms->app_data = NULL; /* Not using app_data */

    return 0;
}

static int
masq_mms_done_1( struct ip_masq_app* pmapp, struct ip_masq* pms )
{
    MOD_DEC_USE_COUNT;

    /* If we were using app_data, we would free it here */

    return 0;
}

/*
 * Handle incoming (S->C) packet
 */
static int
masq_mms_in( struct ip_masq_app* papp, struct ip_masq* pms, struct sk_buff** ppskb, __u32 maddr )
{
    /* Nothing to do here */
    return 0;
}

/*
 * Handle outgoing (C->S) packet
 */
static int
masq_mms_out( struct ip_masq_app* papp, struct ip_masq* pms, struct sk_buff** ppskb, __u32 maddr )
{
    int   diff     = 0;
    char* skbdata  = NULL;
    __u32 skblen   = 0;

    char* pdata    = NULL;
    int   datalen  = 0;

    char* pline;
    __u32 linelen;
    char  buf[32];
    int   buflen;

    __u32 entitylen = 0;
    int isinvite = 0;
    int accepted = 0;
    int addrpos = 0;
    int addrlen = 0;
    int portpos = 0;
    int portlen = 0;
    struct ip_masq* pmsnew;
    __u16 s_port;

    get_skb_data( ppskb, &skbdata, &skblen );
    pdata = skbdata;
    datalen = skblen;

    if( !next_line( &pdata, &datalen, &pline, &linelen ) )
    {
        return 0;
    }

    /* Look for "MSG <seqno> <v> <mimelen>" */
    if( strncmp( pline, "MSG ", 4 ) != 0 ) return 0;
    pline += 4;
    if( ! IS_DIGIT(*pline) ) return 0;
    while( IS_DIGIT(*pline) ) pline++;
    if( *pline != ' ' ) return 0;
    pline++;
    if( ! IS_ALPHA(*pline) ) return 0;
    pline++;
    if( *pline != ' ' ) return 0;
    pline++;
    if( ! IS_DIGIT(*pline) ) return 0;
    pline += str_to_u32( pline, &entitylen );

    /* Parse mime headers */
    while( next_line( &pdata, &datalen, &pline, &linelen ) )
    {
        if( linelen == 0 )
        {
            break;
        }
        if( memcmp( pline, "Content-Type:", 13 ) == 0 )
        {
            pline += 13;
            while( *pline == ' ' ) pline++;
            if( memcmp( pline, "text/x-msmsgsinvite;", 20 ) == 0 )
                isinvite = 1;
        }
    }
    if( !isinvite ) return 0;

    /* Parse entity for accept, ipaddr, port */
    while( next_line( &pdata, &datalen, &pline, &linelen ) )
    {
        if( memcmp( pline, "Invitation-Command:", 19 ) == 0 )
        {
            pline += 19;
            while( *pline == ' ' ) pline++;
            if( memcmp( pline, "ACCEPT", 6 ) == 0 )
                accepted = 1;
        }
        if( memcmp( pline, "IP-Address:", 11 ) == 0 )
        {
            pline += 11;
            while( *pline == ' ' ) pline++;
            addrpos = pline - skbdata;
            addrlen = 0;
            while( IS_DIGITDOT(pline[addrlen]) ) addrlen++;
        }
        if( memcmp( pline, "Port:", 5 ) == 0 )
        {
            pline += 5;
            while( *pline == ' ' ) pline++;
            portpos = pline - skbdata;
            portlen = 0;
            while( IS_DIGIT(pline[portlen]) ) portlen++;
        }
    }

    if( !accepted || !addrpos || !portpos ) return 0;

    str_to_u16( skbdata+portpos, &s_port );

    /* Grab a port */
    pmsnew = ip_masq_new( IPPROTO_TCP,
                          maddr, 0,
                          pms->saddr, htons(s_port),
                          0, 0,
                          IP_MASQ_F_NO_DPORT|IP_MASQ_F_NO_DADDR );
    if( !pmsnew ) return 0;

    /* Replace the data */
    buflen = sprintf( buf, "%d.%d.%d.%d", NIPQUAD(pmsnew->maddr) );
    if( buflen == addrlen )
    {
        memcpy( skbdata+addrpos, buf, buflen );
    }
    else
    {
        *ppskb = ip_masq_skb_replace( *ppskb, GFP_ATOMIC,
                                      skbdata+addrpos, addrlen,
                                      buf, buflen );
        get_skb_data( ppskb, &skbdata, &skblen );
        diff += buflen - addrlen;
        if( portpos > addrpos )
        {
            portpos += buflen - addrlen;
        }
    }
    buflen = sprintf( buf, "%hu", ntohs(pmsnew->mport) );
    if( buflen == portlen )
    {
        memcpy( skbdata+portpos, buf, buflen );
    }
    else
    {
        *ppskb = ip_masq_skb_replace( *ppskb, GFP_ATOMIC,
                                      skbdata+portpos, portlen,
                                      buf, buflen );
        /* no need to call get_skb_data, we are done */
        diff += buflen - portlen;
    }

    /* The operation is a success! */
    ip_masq_listen( pmsnew );
    ip_masq_put( pmsnew );

    return diff;
}

/*** Module load/unload support ***/

static struct ip_masq_app g_masq_functbl =
{
    NULL,               /* next */
    "masq_mms_name",    /* name */
    0,                  /* type */
    0,                  /* n_attach */
    masq_mms_init_1,    /* ip_masq_init_1 */
    masq_mms_done_1,    /* ip_masq_done_1 */
    masq_mms_out,       /* pkt_out */
    masq_mms_in         /* pkt_in */
};

/*
 * Initialization (load) function
 */
static int __init
masq_mms_init( void )
{
    int i;

    memset( g_masq_objs, 0, sizeof(g_masq_objs) );
    for( i = 0; i < MAX_MASQ_APP_PORTS && ports[i]; i++ )
    {
        int s;
        g_masq_objs[i] = kmalloc( sizeof(struct ip_masq_app), GFP_KERNEL );
        if( ! g_masq_objs[i] )
        {
            return -ENOMEM;
        }
        memcpy( g_masq_objs[i], &g_masq_functbl, sizeof(struct ip_masq_app) );
        s = register_ip_masq_app( g_masq_objs[i], IPPROTO_TCP, ports[i] );
        if( s )
        {
            return s;
        }
    }

    return 0;
}

/*
 * Destruction (unload) function
 */
static int
masq_mms_done( void )
{
    int i;
    int rv;

    rv = 0;
    for( i = 0; i < MAX_MASQ_APP_PORTS; i++ )
    {
        if( g_masq_objs[i] )
        {
            int s = unregister_ip_masq_app( g_masq_objs[i] );
            if( s )
            {
                rv = s;
            }
            else
            {
                kfree( g_masq_objs[i] );
                g_masq_objs[i] = NULL;
            }
        }
    }

    return rv;
}

#ifdef MODULE

EXPORT_NO_SYMBOLS;

int init_module( void )
{
    if( masq_mms_init() != 0 )
    {
        return -EIO;
    }

    return 0;
}

void cleanup_module( void )
{
    if( masq_mms_done() != 0 )
    {
        printk( KERN_INFO "ip_masq_mms: can't remove module" );
    }
}

#endif /* MODULE */
